/*
 * Program memory segment support.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>

#include <ipc/mem_ipc.h>
#include <mem/memory_utils.h>

#include <stdio.h>
#include <string.h>

#include "segment.h"



/* Obligatory destructor. */

Segment::~Segment()
{
}

/* Common initialisation. */

void Segment::init()
{
  _region_base = trunc(region_address(), L4_PAGESIZE);
  _region_content_offset = region_address() - _region_base;

  /* Expand the region size. */

  _region_allocated_size = round(region_size() + _region_content_offset, L4_PAGESIZE);
}

/* Return the address of allocated memory. */

char *Segment::address()
{
  return _buf;
}

/* Return the size of the allocated memory. */

offset_t Segment::size()
{
  return _region_allocated_size;
}

/* Allocate a writable region for the segment. */

long Segment::allocate()
{
  return ipc_allocate_align(_region_allocated_size, L4RE_RM_F_SEARCH_ADDR | L4RE_RM_F_RW,
                            L4_PAGESHIFT, (void **) &_buf, &_ds);
}

/* Fill a segment region with file content. */

long Segment::fill(file_t *file)
{
  /* Provide the exposed file contents in a masked memory mapped region. */

  _buf = (char *) client_mmap(file,
                              file_offset() - _region_content_offset,
                              _region_allocated_size,
                              file_offset(),
                              file_offset() + file_contents(),
                              region_flags());

  if (_buf == NULL)
    return -L4_EIO;
  else
    return L4_EOK;
}

/* Define and return the mapped region for the segment. */

MappedRegion &Segment::region()
{
  _region = MappedRegion((l4_addr_t) _buf, _region_allocated_size, region_flags(), _region_base);
  return _region;
}

/* Return the mapped address corresponding to the given address. */

l4_addr_t Segment::region_address(char *address)
{
  return (l4_addr_t) ((address - _buf) + (char *) _region_base);
}

l4_addr_t Segment::region_address(l4_addr_t address)
{
  return (address - (l4_addr_t) _buf) + _region_base;
}



/* Initialise a memory segment using explicit parameters. */

ExplicitSegment::ExplicitSegment(l4_addr_t base, offset_t size,
                                 l4re_rm_flags_t flags, offset_t file_offset,
                                 offset_t file_contents)
: _base(base), _size(size), _flags(flags), _file_offset(file_offset),
  _file_contents(file_contents)
{
  init();
}

/* Return the amount of file content loaded into the segment. */

bool ExplicitSegment::loadable()
{
  return false;
}

offset_t ExplicitSegment::file_contents()
{
  return _file_contents;
}

offset_t ExplicitSegment::file_offset()
{
  return _file_offset;
}

l4_addr_t ExplicitSegment::region_address()
{
  return _base;
}

offset_t ExplicitSegment::region_size()
{
  return _size;
}

/* Return the region flags for the segment. */

l4re_rm_flags_t ExplicitSegment::region_flags()
{
  return _flags;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
